/* ----
 * ---- file   : mlb.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel.
 * ----          Distributed under terms of the MIT LICENSE (MIT).
 * ----
 * ---- Permission is hereby granted, free of charge, to any person obtaining a copy
 * ---- of this software and associated documentation files (the "Software"), to deal
 * ---- in the Software without restriction, including without limitation the rights
 * ---- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * ---- copies of the Software, and to permit persons to whom the Software is
 * ---- furnished to do so, subject to the following conditions:
 * ----
 * ---- The above copyright notice and this permission notice shall be included in
 * ---- all copies or substantial portions of the Software.
 * ----
 * ---- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * ---- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * ---- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * ---- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * ---- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * ---- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * ---- THE SOFTWARE.
 * ----
 * ---- info   : Mailbox handler / component dispatcher. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 10Sep2013, 11Sep2013, 12Sep2013, 13Sep2013, 14Sep2013, 15Sep2013, 17Sep2013
 * ----
 * ----
 */

#define ENABLE_COMPONENT_DISPATCHER defined

// If defined, assume that components are numbered 0..15 (MLB_MAX_COMPONENTS)
//  this is to avoid a linear search in loc_component_find_by_id()
#define ASSUME_LINEAR_COMPONENT_ID_MAPPING defined


#define IVA2_XMC_L1DMPAA_BASE  (0x0184AE00u)


#include <std.h>
#include <log.h>
#include <mem.h>
#include <sem.h>
#include <sys.h>
#include <tsk.h>
#include <hwi.h>
#include <bcache.h>
#include <c64.h>
#include <string.h>  /* for memset() */

#include "../../../include/types.h"
#include "../../../include/arch/dm3730/dm3730.h"
#include "../../../include/dsp_common.h"
#include "../../../include/dsp_config.h"

#include "../include/int.h"
#include "../include/mlb.h"


#define MLB_INT_VEC_ID  IVA2_INT_10


#pragma DATA_ALIGN(mlb_sram,  32);
#pragma DATA_SECTION(mlb_sram,  ".sec_mlb_sram");
unsigned char mlb_sram[DSP_IPC_SIZE];

#define IPC_REG(off)  (*(sU32*)(&mlb_sram[(off)]))

/*
 * it is important to access the L1DSRAM via the compiler "mlb_sram" variable.
 *
 * Access via (*(volatile sU32*)(L1DSRAMADDR)) freezes the DSP (wth!?!)
 */


/* --------------------------------------------------------------------------- vars */
static TSK_Attrs mlb_task_attrs;
static TSK_Handle mlb_task_handle;

//static SEM_Handle mlb_sem_handle;

static volatile sUI num_irqs;
static volatile sUI num_msg_recvd;

static volatile sU32 mlb_inbox[4];

static dsp_component_t components[MLB_MAX_COMPONENTS];
static sUI num_components;


/* --------------------------------------------------------------------------- mlb component (0) */
static sU32 loc_mlb_exec(dsp_component_cmd_t _cmd,
                         sU32  _arg1, sU32  _arg2,
                         sU32 *_ret1, sU32 *_ret2
                         ) {
  sU32 ret;

  /* Succeeded */
  ret = 0;

  return ret;
}


static dsp_component_t component_mlb = {
      0, /* component id, assigned in mlb_component_register() */

      /* fxns: */
      {
         NULL, /* init */
         &loc_mlb_exec,
         NULL  /* exit */
      }
};


/* --------------------------------------------------------------------------- loc_debug() */
void loc_debug(sU32 _val) {

   IPC_REG(DSP_IPC_DEBUG_INTERNAL_OFF) = _val;

   ////BCACHE_wbInv((void*)DSP_MLB_RAM_DSPR, 4, 1/*wait*/);
}


/* --------------------------------------------------------------------------- loc_component_find_by_id() */
dsp_component_t *loc_component_find_by_id(dsp_component_id_t _id) {
   sUI i;
   dsp_component_t *ret = components;

#ifdef ASSUME_LINEAR_COMPONENT_ID_MAPPING
   if(_id < num_components)
   {
      ret += _id;
   }
   else
   {
      ret = NULL;
   }
#else

   /* (todo) optimize */
   for(i=0; i<num_components; i++)
   {
      if(ret->component_id == _id)
      {
         /* Found component */
         break;
      }

      /* Try next component */
      ret++;
   }

   if(MLB_MAX_COMPONENTS == i)
   {
      /* Failed: component not found */
      ret = NULL;
   }

#endif /* ASSUME_LINEAR_COMPONENT_ID_MAPPING */
   return ret;
}


/* --------------------------------------------------------------------------- loc_its_alive() */
void loc_its_alive(void) {
   /* Write some identity values so that GPP side can verify that DSP is actually running
    *
    *  (note) the first two cachelines (64 bytes) are reserved for GPP<>DSP communication
    *
    */
   sU32 i;
   volatile sU32 *resetBuf = (sU32*) (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE);

   for(i=0; i<16; i++)
   {
      resetBuf[i] = 0;
   }

   for(i=16; i<((DSP_RESET_VECTOR_MINSIZE - DSP_RESET_VECTOR_CODESIZE)/4); i++)
   {
      resetBuf[i] = i;
   }

   BCACHE_wbAll();
}


/* --------------------------------------------------------------------------- mlb_irq__handler() */
void mlb_irq_handler(void *_arg) {
   (void)_arg;

   /* Acknowledge interrupt */
   DSP_REG(MLB_IRQSTATUS_1) = (1u << 0); // clear NEWMSGSTATUSUUMB0

   // Disable mailbox interrupt to prevent interrupt spam
   //  (workaround for buggy hardware)
   DSP_REG(MLB_IRQENABLE_1) = 0;

#if 1
   {
      sU32 numMsg = DSP_REG(MLB_MSGSTATUS_0) & 7u;

      if(numMsg > 0)
      {
         sUI fifoIdx = 0;

         while(fifoIdx < numMsg)
         {
            mlb_inbox[fifoIdx] = DSP_REG(MLB_MESSAGE_0);

            // Next message
            fifoIdx++;
         }
      }
   }
#endif

   int_irq_clear(MLB_INT_VEC_ID);

   int_event_clear(IVA2_EVT_MAIL_U1_IRQ);

   C64_clearIFR(1 << MLB_INT_VEC_ID);

   //SEM_post(mlb_sem_handle);

   BCACHE_wbInvAll();

   num_irqs++;
}


/* --------------------------------------------------------------------------- mlb_task__entry() */
void mlb_send_to_gpp(sU32 _msg) {

   DSP_REG(MLB_MESSAGE_1) = ~mlb_inbox[0];
}


/* --------------------------------------------------------------------------- mlb_task__entry() */
static void mlb_task__entry(void) {

   num_irqs      = 0;
   num_msg_recvd = 0;

   //mlb_sem_handle = SEM_create(0, NULL);



   /* Check whether MAILBOX registers can be accessed
    *
    *  (note) the "c64.ko" kernel module must have been loaded successfully on the GPP side
    */
   /////if(0 == DSP_REG(CORE_CM_IDLEST1) & 0x80)  // does not work on DSP side, apparently !??!
   {
      ////sBool bRecvd = 0;
      sU32 iterCount = 0x10000000u;

#if 1
      HWI_disable();

      /* Map MAIL_U1_IVA2_IRQ (55) to mailbox interrupt vector (10) */
      HWI_eventMap(
            MLB_INT_VEC_ID,
            IVA2_EVT_MAIL_U1_IRQ
            );

      /* Add interrupt handler to HWI dispatcher table */
      HWI_dispatchPlug(
            MLB_INT_VEC_ID,       // "vecid must be a valid interrupt ID in the range of 0-15."
            (Fxn)mlb_irq_handler,
            -1,                   // dma channel, -1 == use CPU copy
            NULL                  // use default args
            );

      /* Enable event source */
      //int_event_enable(IVA2_EVT_MAIL_U1_IRQ);

      /* Unmask mailbox interrupt (IVA2_IRQ[10]) */
      ////HWI_enableWugen(10);
      DSP_REG(IVA2_WUGEN_MEVTCLR0) = (1u << 10);

      num_irqs = 0x20000000u;

      HWI_enable();

      /* Setup mailbox 0 for receiving messages from the DSP to the GPP, and
       *       mailbox 1 for sending messages to the GPP
       */
      DSP_REG(MLB_IRQENABLE_1) =
            (
               //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
               (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
             );

      /* Enable mailbox interrupt */
      C64_clearIFR(1 << MLB_INT_VEC_ID);//C64_EINT4);
      C64_enableIER(1 << MLB_INT_VEC_ID);//C64_EINT4);
#endif

#if 1
      {
         sUI i = 0;

         while(i < 1) // (note) crash when writing to all 32 regs (or even more than 2)
         {
            /* Unlock access to L1D memory */
            DSP_REG(IVA2_XMC_L1DMPAA_BASE + (4u * i)) =
                  (1u <<  1) |  // User write access is permitted
                  (1u <<  2) |  // User read access is permitted
                  (1u <<  4) |  // Supervisor write access is permitted
                  (1u <<  5) |  // Supervisor read access is permitted
                  (1u <<  8) |  // Local (DSP) access is permitted
#if 1
                  (1u <<  9) |  // External access is permitted
                  (1u << 10) |  // ID0 has access permission
                  (1u << 11) |  // ID1 has access permission
                  (1u << 12) |  // ID2 has access permission
                  (1u << 13) |  // ID3 has access permission
                  (1u << 14) |  // ID4 has access permission
                  (1u << 15) |  // ID5 has access permission
#endif
                  0          ;

            i++;
         }
      }
#endif

      ////BCACHE_wbInvAll();

      if(1)
      {
         sUI i;

         for(i=0; i<64; i++)
         {
            mlb_sram[i] = i;
         }
      }

      if(1)
      {
         sUI i;

         mlb_inbox[3] = 0;

         for(i=0; i<64; i++)
         {
            mlb_inbox[3] += mlb_sram[i];
         }
      }

      loc_debug(0x07000001u);


      /* Fill buffer so ARM-side can see that the DSP is doing something */
      loc_its_alive();

      ////for(;;);


      loc_debug(0x07000002u);

      mlb_debug_usr(0, 0xCCddCCddu);

      {
         sUI oldNumIrqs = num_irqs;
         sUI oldMsgSerial = 0;

         for(;;)
         {
            if(num_irqs != oldNumIrqs)
            {
               sUI newMsgSerial;

               loc_debug(0x07000003u);

               oldNumIrqs = num_irqs;

               ////BCACHE_inv((Ptr)DSP_MLB_RAM_MSGSERIAL, 32, 1/*wait*/);

               newMsgSerial = IPC_REG(DSP_IPC_MSG_SERIAL_OFF);

               IPC_REG(DSP_IPC_DEBUG_DSPIRQCOUNT_OFF) = num_irqs;

               loc_debug(0x07000004u);

               ////BCACHE_wbInv((void*)DSP_MLB_RAM_DSPIRQCOUNT, 8, 1/*wait*/);

               if(oldMsgSerial != newMsgSerial)
               {
                  sU32 rpcDebug = 0;

                  oldMsgSerial = newMsgSerial;

                  loc_debug(0x07000005u);

                  /* Try to execute requested command */
#ifdef ENABLE_COMPONENT_DISPATCHER
                  {
                     sU32 compAndCmd = IPC_REG(DSP_IPC_MSG_COMP_OFF);

                     const dsp_component_t *comp = loc_component_find_by_id(
                           (dsp_component_id_t)(compAndCmd & 0xFFFFu)
                           );

                     if(NULL != comp)
                     {
                        sU32 err;

                        /* Execute component command
                         *
                         *  (note) the first +4u skips the "client" field of the dsp_msg_t structure
                         *  (note) +2u then addresses the cmd_id (sU16)
                         *  (note) the remaining +4u resp. +8u address the arg1 resp. arg2 fields
                         *
                         *  (note) also see "msg.h"
                         */

                        sU32 r1;
                        sU32 r2;

                        err = comp->fxns.exec(
                              (dsp_component_cmd_t)(compAndCmd >> 16u),
                              IPC_REG(DSP_IPC_MSG_ARG1_OFF),
                              IPC_REG(DSP_IPC_MSG_ARG2_OFF),
                              &r1,
                              &r2
                              );

                        IPC_REG(DSP_IPC_MSG_RET1_OFF) = r1;
                        IPC_REG(DSP_IPC_MSG_RET2_OFF) = r2;

                        IPC_REG(DSP_IPC_ERRNO_OFF) = err;

                        /////BCACHE_wbInv((void*)DSP_MLB_RAM_MSGTOGPP, 32, 1/*wait*/);

                        if(0 == err)
                        {
                           /* Ok, command completed successfully */
                           rpcDebug = 0x70000000u;

                           rpcDebug |= (r1 & 255) << 8u;
                        }
                        else
                        {
                           /* (todo) propagate error code to GPP */
                           rpcDebug = 0x30000000u;

                           rpcDebug |=
                                 ((err & 255u) << 16u) |
                                 ((*(dsp_component_cmd_t*)(IPC_REG(DSP_IPC_MSG_COMP_OFF))) << 8u)
                                 ;
                        }

                     }
                     else
                     {
                        /* Failed: component not found */
                        rpcDebug = 0x10000000u;
                     }
                  }
#endif /* ENABLE_COMPONENT_DISPATCHER */


                  /* Write back caches so GPP can see the data */
                  //BCACHE_wbInvAll();

                  loc_debug(0x07000006u | rpcDebug);

                  /* Send reply to trigger IRQ on GPP-side.
                   *
                   *  (note) the content of the message register does not matter,
                   *          it's only written to in order to trigger an interrupt.
                   *
                   *  (note) even if the command failed or the requested component does not exist,
                   *          a reply message must be sent.
                   */
                  mlb_send_to_gpp(~mlb_inbox[0]);

                  loc_debug(0x07000007u | rpcDebug);

                  iterCount = 0;

               } // if(oldMsgSerial != newMsgSerial)

            } // if(num_irqs != oldNumIrqs)

            iterCount++;

            //if(0 == (iterCount & 1023))
            {
               IPC_REG(DSP_IPC_DEBUG_DSPITERCOUNT_OFF) = iterCount;
            }

            ////TSK_sleep(1); // never returns

         } // for(;;);
      }
   }
}


/* --------------------------------------------------------------------------- mlb_init() */
void mlb_init(void) {
   /* (note) must be called in main() before registering components */

   memset(components, 0, sizeof(components));

   /* Register the MLB component */
   components[0] = component_mlb;
   components[0].component_id = 0;

   num_components = 1;
}


/* --------------------------------------------------------------------------- mlb_start() */
sBool mlb_component_register (dsp_component_id_t     _compId,
                              const dsp_component_t *_comp
                              ) {
   sBool ret;

   if(_compId > 0)
   {
      if(NULL != _comp)
      {
         if(num_components < MLB_MAX_COMPONENTS)
         {
            if(NULL != _comp->fxns.exec)
            {
               components[num_components] = *_comp;

               components[num_components].component_id = _compId;

               num_components++;

               /* Succeeded */
               ret = S_TRUE;
            }
            else
            {
               /* Failed: no exec function */
               ret = S_FALSE;
            }
         }
         else
         {
            /* Failed: maximum number of components exceeded */
            ret = S_FALSE;
         }
      }
      else
      {
         /* Failed: component reference is NULL */
         ret = S_FALSE;
      }
   }
   else
   {
      /* Failed: illegal component id */
      ret = S_FALSE;
   }

   return ret;
}


/* --------------------------------------------------------------------------- mlb_start() */
void mlb_start(void) {

   /* Initialize components */
   {
      sUI i;

      for(i=0; i<num_components; i++)
      {
         if(NULL != components[i].fxns.init)
         {
            if(!components[i].fxns.init())
            {
               /* Failed: component failed to initialize */
               mlb_panic();
            }
         }

         /* Next component */
      }
   }

   /* Create ARM<>DSP mailbox task
    *
    *  (todo) use tconf to create the task statically (size opt.)
    */
   mlb_task_attrs = TSK_ATTRS;
   mlb_task_attrs.name     = "mlb";
   mlb_task_attrs.priority = TSK_MINPRI;

   mlb_task_handle = TSK_create((Fxn)mlb_task__entry, &mlb_task_attrs);
}


/* --------------------------------------------------------------------------- mlb_debug_usr() */
void mlb_debug_usr(sU32 _idx, sU32 _val) {

   IPC_REG(DSP_IPC_DEBUG_USR0_OFF + (4u * (_idx & 3u))) = _val;

   ////BCACHE_wbInv((void*)DSP_MLB_RAM_DSPDEBUG, 4, 1/*wait*/);
}


/* --------------------------------------------------------------------------- mlb_panic() */
void mlb_panic(void) {

   /* (todo) signal GPP-side that it needs to reboot the DSP */

   for(;;)
   {
      /* Loop a while.. loop..forever! */

      /* (todo) execute idle instruction ?! */
   }
}
