/* ----
 * ---- file   : dev.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 11Sep2013, 12Sep2013, 14Sep2013, 16Sep2013, 17Sep2013, 18Sep2013
 * ----
 * ----
 */

//#define USE_DRIVER defined

#define __NO_VERSION__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/string.h>
#include <linux/dma-mapping.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>

#include <linux/../../arch/arm/include/asm/cacheflush.h>

#include "../include/types.h"
#include "../include/msg.h"
#include "../include/dsp_common.h"
#include "../include/c64_dev.h"

#include "kmod.h"
#include "reg.h"
#include "mlb.h"
#include "dev.h"


/*--------------------------------------------------------------------------- module vars */
#ifdef USE_DRIVER
static struct device_driver c64_driver = {
   .name = "c64",
   .bus  = &platform_bus_type /* ?!? */
};
#endif /* USE_DRIVER */

static dev_t          c64_dev_nr;
static struct cdev *  c64_cdev;
static struct class * c64_class;
static struct device *c64_device;


/*--------------------------------------------------------------------------- client typedefs */
typedef enum {
   STATE_UNUSED        = 0,  // client slot is free for allocation (open).
   STATE_IDLE          = 1,  // client is connected but there's no activity ATM.
   STATE_REQ_QUEUED    = 2,  // (todo) message is queued and will be sent to DSP ASAP.
   STATE_REPLY_PENDING = 3,  // waiting for reply from DSP.
   STATE_REPLY_AVAIL   = 4,  // reply from DSP has been recv'd but not picked up by the client, yet.
} dev_client_state_t;


typedef struct {
   atomic_t         at_state;
   struct file *    filep;
   c64_msg_t        c64_msg;
   dsp_mem_region_t shm;
   void *           shm_kvirt_addr;

} dev_client_t;


/*--------------------------------------------------------------------------- client module vars */
static dev_client_t clients[C64_DEV_MAX_CLIENTS];

static volatile dev_client_t *current_client; // != NULL while transmission in progress


/*--------------------------------------------------------------------------- loc_init_clients() */
static void loc_init_clients(void) {
   /* Initialize clients array */
   sUI i;
   dev_client_t *cl = clients;
   
   for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
   {
      atomic_t atInit = ATOMIC_INIT(STATE_UNUSED);
      
      memset(cl, 0, sizeof(dev_client_t));
      
      cl->at_state = atInit; /* (todo) is this allowed ? */
      
      /* Next client */
      cl++;
   }
}


/*--------------------------------------------------------------------------- loc_find_client_by_file() */
static dev_client_t *loc_find_client_by_filep(struct file *_filep) {
   dev_client_t *ret;
   sUI i;

   ret = clients;

   for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
   {
      if(ret->filep == _filep)
      {
         if(STATE_UNUSED != atomic_read(&ret->at_state))
         {
            /* Succeeded */
            break;
         }
      }

      /* Try next client struct */
      ret++;
   }

   if(C64_DEV_MAX_CLIENTS == i)
   {
      /* Failed: client not found */
      ret = NULL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- loc_shm_alloc() */
long loc_shm_alloc(dev_client_t *_client, c64_dev_ioctl_shm_alloc_t *_req) {
   long ret;
   sUI numBytes;
   dma_addr_t dmaAddr;
   void *kVirtAddr;

   /* Round up req.size to page size 
      xxx hardcoded pagesize
   */
   numBytes = (((_req->mem.size + 4095u) >> 12u) << 12u);

   if(C64_DEBUG_LEVEL(10))
   {
      printk(KERN_DEBUG "c64::dev::loc_shm_alloc: calling dma_alloc_coherent()\n");
   }
   
   
   /* Allocate physically contiguous, uncached shared memory */
   kVirtAddr = dma_alloc_coherent(NULL,//c64_device,
                                  (size_t)numBytes,
                                  //(dma_addr_t*)&_client->shm.phys_addr,
                                  &dmaAddr,
                                  GFP_KERNEL
                                  );
   
   if(C64_DEBUG_LEVEL(10))
   {
      printk(KERN_DEBUG "c64::dev::loc_shm_alloc: dma_alloc_coherent() returned kVirtAddr=0x%08x physAddr=0x%08x\n",
             (sU32)kVirtAddr,
             (sU32)dmaAddr
             );
   }
   
   if(NULL != kVirtAddr)
   {
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_INFO "c64::dev::loc_shm_alloc: allocated %u kBytes (%u pages).\n",
                (_req->mem.size >> 10u),
                (_req->mem.size >> 12u)
                );
      }

      _client->shm.size       = (sU32) _req->mem.size;
      _client->shm.phys_addr  = (sU32) dmaAddr;
      _client->shm.virt_addr  = (sU32) 0; /* will be set in fops__mmap() */
      _client->shm_kvirt_addr = kVirtAddr;

      /* Succeeded */
      ret = 0;
   }
   else
   {
      /* Failed: dma_alloc_coherent() */
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_ERR "c64::dev::ioctl::loc_shm_alloc: failed to allocate %u kBytes (%u pages).\n",
                (_req->mem.size >> 10u),
                (_req->mem.size >> 12u)
                );
      }

      ret = -ENOMEM;
   }

   return ret;
}


/*--------------------------------------------------------------------------- loc_shm_free() */
long loc_shm_free(dev_client_t *_client) {
   long ret;

   if(NULL != _client->shm_kvirt_addr)
   {
      if(C64_DEBUG_LEVEL(10))
      {
         printk(KERN_DEBUG "c64::dev::loc_shm_free: calling dma_free_coherent()\n");
      }

      dma_free_coherent(NULL,//c64_device,
                        _client->shm.size,
                        (void*)_client->shm_kvirt_addr,
                        (dma_addr_t)_client->shm.phys_addr
                        );

      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_INFO "c64::dev::loc_shm_free: freed %u kBytes (%u pages).\n",
                (_client->shm.size >> 10u),
                (_client->shm.size >> 12u)
                );
      }

      _client->shm_kvirt_addr = NULL;
      _client->shm.size       = 0;
      _client->shm.virt_addr  = 0;
      _client->shm.phys_addr  = 0;
                        
      /* Succeeded */
      ret = 0;
   }
   else
   {
      /* Nothing's allocated ATM */
      ret = 0;
   }

   return ret;
}


/*--------------------------------------------------------------------------- FOPS module vars */
static int     fops__open     (struct inode *_inode, struct file *_filep);
static int     fops__release  (struct inode *_inode, struct file *_filep);
static ssize_t fops__read     (struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos);
static ssize_t fops__write    (struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos);
static long    fops__unlocked_ioctl (struct file *_filep, unsigned int _cmd, unsigned long _arg);
static int     fops__mmap     (struct file *_filep, struct vm_area_struct *_vma);

#define FOP(a) .a = fops__##a

static struct file_operations fops = {
   .owner = THIS_MODULE,

   FOP(open),
   FOP(release),
   FOP(read),
   FOP(write),
   FOP(unlocked_ioctl),
   FOP(mmap)
};

/* (note) may temporarily exceed C64_DEV_MAX_CLIENTS */
static atomic_t at_num_open = ATOMIC_INIT(0);


/*--------------------------------------------------------------------------- fops__open */
int fops__open(struct inode *_inode, struct file *_filep) {
   int ret;

   if(C64_DEV_MINOR == iminor(_inode))
   {
      if(atomic_inc_and_test(&at_num_open) < C64_DEV_MAX_CLIENTS)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL == cl)
         {
            /* Find first unused client */
            sUI i;
            cl = clients;

            for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
            {
               if(STATE_UNUSED == atomic_read(&cl->at_state))
               {
                  /* Found unused client struct */
                  cl->filep = _filep;
                  atomic_set(&cl->at_state, STATE_IDLE);
                  break;
               }
               
               /* Try next client struct */
               cl++;
            }

            if(C64_DEV_MAX_CLIENTS != i)
            {
               /* Succeeded */
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_INFO "c64::dev::open: client connected (filep=0x%p).\n", _filep);
               }

               ret = 0;
            }
            else
            {
               /* Failed: all clients in use */
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_ERR "c64::dev::open: all clients in use.\n");
               }

               ret = -EMFILE;
            }
         }
         else
         {
            /* Failed: client not found (internal error) */
            printk(KERN_ERR "c64::dev::open: internal error: client (filep=0x%p) already exists.\n", _filep);

            ret = -EIO;
         }
      }
      else
      {
         /* Failed: maximum number of file handles exceeded */
         atomic_dec(&at_num_open);

         ret = -EMFILE;
      }
   }
   else
   {
      /* Failed: wrong minor device number */
      ret = -ENODEV;
      
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__release */
int fops__release(struct inode *_inode, struct file *_filep) {
   int ret;
   dev_client_t *cl;

   atomic_dec(&at_num_open);

   cl = loc_find_client_by_filep(_filep);

   if(NULL != cl)
   {
      /* Free shared memory in case client forgot to */
      if(0/*NULL*/ != cl->shm.phys_addr)
      {
         loc_shm_free(cl);
      }

      cl->filep = NULL;
      atomic_set(&cl->at_state, STATE_UNUSED);

      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_INFO "c64::dev::release: client disconnected (filep=0x%p).\n", _filep);
      }
   }
   else
   {
      /* Failed: client not found (internal error) */
      printk(KERN_ERR "c64::dev::release: internal error: client (filep=0x%p) not found.\n", _filep);
   }

   /* Succeeded */
   ret = 0;

   return ret;
}


/*--------------------------------------------------------------------------- fops__read */
ssize_t fops__read(struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   (void)_pos;

   if(C64_DEBUG_LEVEL(30))
   {
      printk(KERN_INFO "c64::dev::read: ENTER  count=%u\n", _count);
   }

   if(sizeof(c64_msg_t) == _count)
   {
      dev_client_t *cl = loc_find_client_by_filep(_filep);

      if(NULL != cl)
      {
         dev_client_state_t cState = atomic_read(&cl->at_state);

         if(STATE_REPLY_AVAIL == cState)
         {
            if(0 == copy_to_user(_buf, &cl->c64_msg, sizeof(c64_msg_t)))
            {
               /* Succeeded */
               if(C64_DEBUG_LEVEL(15))
               {
                  printk(KERN_INFO "c64::dev:read: DSP reply picked up by application.\n");
               }

               atomic_set(&cl->at_state, STATE_IDLE);

               ret = sizeof(c64_msg_t);
            }
            else
            {
               /* Failed: copy_to_user() */
               printk(KERN_ERR "c64::dev::read: copy_to_user() failed.\n");

               return -EFAULT;
            }
         }
         else
         {
            if(STATE_REPLY_PENDING == cState)
            {
               /* Failed: no reply available, yet. */
               if(C64_DEBUG_LEVEL(30))
               {
                  printk(KERN_ERR "c64::dev::read: reply still pending, cState=%d.\n", cState);
               }

               ret = -EAGAIN;
            }
            else
            {
               /* Failed: (warning) nothing to read. */
               if(C64_DEBUG_LEVEL(15))
               {
                  printk(KERN_ERR "c64::dev::read: nothing to read (no pending reply).\n");
               }

               ret = 0;
            }
         }
      }
      else
      {
         /* Failed: client not found */
         if(C64_DEBUG_LEVEL(1))
         {
            printk(KERN_ERR "c64::dev::read: client (filep=0x%p) not found.\n", _filep);
         }

         ret = -EBADF;
      }
   }
   else if(4 == _count)
   {
      /* Read debug value */
      sU32 val = mlb_debug_read();

      if(0 == copy_to_user(_buf, &val, 4))
      {
         /* Succeeded */
         ret = 4;
      }
      else
      {
         /* Failed: copy_to_user() */
         return -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_ERR "c64::dev::read: invalid size %d.\n", _count);
      }

      ret = -EINVAL;
   }

   if(C64_DEBUG_LEVEL(30))
   {
      printk(KERN_INFO "c64::dev::read: LEAVE  ret=%d\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__write */
ssize_t fops__write(struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   if(sizeof(c64_msg_t) == _count)
   {
      if(NULL == current_client)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL != cl)
         {
            if(0 == copy_from_user(&cl->c64_msg, _buf, sizeof(c64_msg_t)))
            {
               current_client = cl;
               
               atomic_set(&cl->at_state, STATE_REPLY_PENDING);

               ret = (ssize_t) mlb_send_to_dsp(cl, &cl->c64_msg);
               
               if(0 == ret)
               {
                  /* Succeeded */
                  ret = sizeof(c64_msg_t);
               }
               else
               {
                  /* Failed: mlb_send_to_dsp() */
                  if(C64_DEBUG_LEVEL(15))
                  {
                     printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (still busy).\n");
                  }
                  
                  current_client = NULL;

                  atomic_set(&cl->at_state, STATE_IDLE);
                  
                  ret = -EAGAIN;
               }
            }
            else
            {
               /* Failed: copy_from_user() */
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_ERR "c64::dev::write: copy_from_user() failed.\n");
               }
               
               ret = -EFAULT;
            }
         }
         else
         {
            /* Failed: client not found */
            if(C64_DEBUG_LEVEL(1))
            {
               printk(KERN_ERR "c64::dev::write: client (filep=0x%p) not found.\n", _filep);
            }
            
            ret = -EBADF;
         }
      }
      else
      {
         /* Failed: transmission still in progress (current_client != NULL) */
         if(C64_DEBUG_LEVEL(30))
         {
            printk(KERN_ERR "c64::dev::write: (current_client != NULL) (still busy).\n");
         }
         
         ret = -EAGAIN;
      }
   }
   else if(4 == _count)
   {
      /* Send test message to DSP, bypassing client handler */
      sU32 val;

      if(0 == copy_from_user(&val, _buf, 4))
      {
         current_client = NULL;

         ret = (ssize_t) mlb_send_to_dsp(NULL, (const c64_msg_t*) &val);

         if(0 == ret)
         {
            ret = 4;
         }
         else
         {
            /* Failed: mlb_send_to_dsp() */
            if(C64_DEBUG_LEVEL(1))
            {
               printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (testmsg).\n");
            }
         }
      }
      else
      {
         /* Failed: copy_from_user() */
         if(C64_DEBUG_LEVEL(1))
         {
            printk(KERN_ERR "c64::dev::write: copy_from_user() (4) failed.\n");
         }
         
         ret = -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_ERR "c64::dev::write: invalid size %d.\n", _count);
      }
      
      ret = -EINVAL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__unlocked_ioctl() */
long fops__unlocked_ioctl(struct file *_filep, unsigned int _cmd, unsigned long _arg) {
   long ret;
   dev_client_t *cl;

   cl = loc_find_client_by_filep(_filep);

   if(NULL != cl)
   {
      switch(_cmd)
      {
         default:
            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_DEBUG "c64::dev::ioctl: illegal cmd=%u\n", _cmd);
            }

            ret = -EINVAL;
            break;
            
         case C64_IOCTL_SHM_ALLOC:
            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_DEBUG "c64::dev::ioctl: cmd=SHM_ALLOC\n");
            }

#if 1
            if(0/*NULL*/ == cl->shm.phys_addr)
            {
               c64_dev_ioctl_shm_alloc_t req;

               if(0 == copy_from_user(&req, (const void __user *)_arg, sizeof(c64_dev_ioctl_shm_alloc_t)))
               {
                  ret = loc_shm_alloc(cl, &req);

                  if(0 == ret)
                  {
                     /* Copy dsp_mem_region_t struct */
                     req.mem = cl->shm;

                     if(0 == copy_to_user((void __user *)_arg, &req, sizeof(c64_dev_ioctl_shm_alloc_t)))
                     {
                        /* Succeeded */
                     }
                     else
                     {
                        /* Failed: copy_to_user() */
                        loc_shm_free(cl);
                        
                        ret = -EFAULT;
                     }
                  }
               }
               else
               {
                  if(C64_DEBUG_LEVEL(1))
                  {
                     printk(KERN_ERR "c64::dev::ioctl: SHM_ALLOC: copy_from_user() failed.\n");
                  }
               
                  ret = -EFAULT;
               }
            }
            else
            {
               /* Only one shared memory block allowed per client */
               ret = -EBUSY;
            }
#endif
            ret = 0;

            break;
            
         case C64_IOCTL_SHM_FREE:
            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_DEBUG "c64::dev::ioctl: cmd=SHM_FREE\n");
            }

#if 1
            ret = loc_shm_free(cl);
#endif
            ret = 0;

            break;

         case C64_IOCTL_TEST:
            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_INFO "c64::dev::ioctl: cmd=TEST\n");
            }

#if 1
            // Try to write to L1DSRAM (works)
            printk(KERN_INFO "c64::dev::ioctl: write L1DSRAM:\n");
            reg32_write(0x5cf04000u, 0x12345678u);

            printk(KERN_INFO "c64::dev::ioctl: read L1DSRAM:\n");
            printk(KERN_INFO "  ==> val=0x%08x\n", reg32_read(0x5cf04000u));
#endif

            ret = 0;
            break;

         case C64_IOCTL_CACHE:
            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_INFO "c64::dev::ioctl: cmd=CACHE\n");
            }

            if(0/*NULL*/ != cl->shm.virt_addr)
            {
               c64_dev_ioctl_cache_t req;

               if(0 == copy_from_user(&req, (const void __user *)_arg, sizeof(c64_dev_ioctl_cache_t)))
               {
                  if((req.phys_addr >= cl->shm.phys_addr)                  &&
                     (req.phys_addr <  (cl->shm.phys_addr + cl->shm.size)) && /* b.c. of possible 32bit overflow */
                     ((req.phys_addr + req.size) <= (cl->shm.phys_addr + cl->shm.size))
                     )
                  {
                     switch(req.action)
                     {
                        default:
                           /* Failed: unsupported action */
                           if(C64_DEBUG_LEVEL(1))
                           {
                              printk(KERN_ERR "c64::dev::ioctl: CACHE: unsupported action (%u).\n", req.action);
                           }
                           
                           ret = -EINVAL;
                           break;
                           
                        case C64_CACHE_AC_WB:
                           /* This is loosely borrowed from TI's CMEM (thanks!) */
                           dmac_map_area(cl->shm_kvirt_addr + (req.phys_addr - cl->shm.phys_addr),
                                         req.size,
                                         DMA_TO_DEVICE
                                         );

                           outer_clean_range(req.phys_addr,
                                             req.phys_addr + req.size
                                             );

                           if(C64_DEBUG_LEVEL(20))
                           {
                              printk(KERN_INFO "c64::dev::ioctl: cache writeback (start=0x%08x size=0x%08x) succeeded.\n",
                                     req.phys_addr,
                                     req.size
                                     );
                           }

                           /* Succeeded */
                           ret = 0;
                           break;
                           
                        case C64_CACHE_AC_INV:
                           /* This is loosely borrowed from TI's CMEM (thanks!) */
                           dmac_map_area(cl->shm_kvirt_addr + (req.phys_addr - cl->shm.phys_addr),
                                         req.size,
                                         DMA_FROM_DEVICE
                                         );

                           outer_inv_range(req.phys_addr,
                                           req.phys_addr + req.size
                                           );

                           if(C64_DEBUG_LEVEL(20))
                           {
                              printk(KERN_INFO "c64::dev::ioctl: cache invalidate (start=0x%08x size=0x%08x) succeeded.\n",
                                     req.phys_addr,
                                     req.size
                                     );
                           }

                           /* Succeeded */
                           ret = 0;
                           break;
                           
                        case C64_CACHE_AC_WBINV:
                           /* This is loosely borrowed from TI's CMEM (thanks!) */
                           dmac_map_area(cl->shm_kvirt_addr + (req.phys_addr - cl->shm.phys_addr),
                                         req.size,
                                         DMA_BIDIRECTIONAL
                                         );

                           outer_flush_range(req.phys_addr,
                                             req.phys_addr + req.size
                                             );

                           if(C64_DEBUG_LEVEL(20))
                           {
                              printk(KERN_INFO "c64::dev::ioctl: cache writeback (start=0x%08x size=0x%08x) succeeded.\n",
                                     req.phys_addr,
                                     req.size
                                     );
                           }

                           /* Succeeded */
                           ret = 0;
                           break;
                     }
                  }
                  else
                  {
                     /* Failed: requested address range is outside of CMA block area */
                     if(C64_DEBUG_LEVEL(1))
                     {
                        printk(KERN_ERR "c64::dev::ioctl: CACHE: requested phys_addr=0x%08x, size=0x%08x outside of CMA area.\n",
                               req.phys_addr,
                               req.size
                               );
                     }

                     ret = -EFAULT;
                  }
               }
               else
               {
                  /* Failed: copy_from_user() */
                  if(C64_DEBUG_LEVEL(1))
                  {
                     printk(KERN_ERR "c64::dev::ioctl: CACHE: copy_from_user() failed.\n");
                  }
               
                  ret = -EFAULT;
               }
            }
            else
            {
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_ERR "c64::dev::ioctl: CACHE: pls call C64_IOCTL_SHM_ALLOC and mmap() first.\n");
               }

               ret = -EAGAIN;
            }
            break;

         case C64_IOCTL_DEBUG_USR:
         {
            c64_dev_ioctl_debug_usr_t req;

            if(C64_DEBUG_LEVEL(10))
            {
               printk(KERN_INFO "c64::dev::ioctl: cmd=DEBUG_USR\n");
            }

            if(0 == copy_from_user(&req, (const void __user *)_arg, sizeof(c64_dev_ioctl_debug_usr_t)))
            {
               sUI idx = req.idx;

               if(idx < 4u)
               {
                  req.val = mlb_debug_usr_read(idx);

                  if(C64_DEBUG_LEVEL(10))
                  {
                     printk(KERN_INFO "c64::dev::ioctl: read user debug[%u] val=0x%08x\n",
                            idx,
                            req.val
                            );

                  }

                  if(0 == copy_to_user((void __user *)_arg, &req, sizeof(c64_dev_ioctl_debug_usr_t)))
                  {
                     /* Succeeded */
                     ret = 0;
                  }
                  else
                  {
                     /* Failed: copy_to_user() */
                     printk(KERN_ERR "c64::dev::ioctl: copy_to_user() (cmd=CACHE) failed.\n");
                     
                     return -EFAULT;
                  }
               }
               else
               {
                  /* Failed: debug index out of range (0..3) */
                  if(C64_DEBUG_LEVEL(1))
                  {
                     printk(KERN_ERR "c64::dev::ioctl: illegal user debug idx (%u)\n", idx);
                  }

                  ret = -EINVAL;
               }
            }
            else
            {
               /* Failed: copy_from_user() */
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_ERR "c64::dev::ioctl: CACHE: copy_from_user() failed.\n");
               }
               
               ret = -EFAULT;
            }
         }
         break;

      } /* switch(_cmd) */
   }
   else
   {
      /* Failed: client not found ?!?? */
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_ERR "c64::dev::ioctl: client (filep=0x%p) not found.\n", _filep);
      }

      ret = -EBADF;
   }

   if(C64_DEBUG_LEVEL(10))
   {
      printk(KERN_ERR "c64::dev::ioctl: LEAVE ret=%d\n", (sSI)ret);
   }
      
   return ret;
}


/*--------------------------------------------------------------------------- fops__mmap() */
static int fops__mmap(struct file *_filep, struct vm_area_struct *_vma) {
   /*
    * (note) I had a peek at how TI's CMEM kernel module (GPL) implements this
    *         (saved me some time, thanks guys'n'gals!)
    *
    * This <http://www.cs.binghamton.edu/~ghose/CS529/sharedmemmap.htm> also has some nice explanations.
    *
    */
   dev_client_t *cl;
   size_t reqSize;
   int ret;

   cl = loc_find_client_by_filep(_filep);

   if(NULL != cl)
   {
      if(0/*NULL*/ == cl->shm.virt_addr)
      {
         reqSize = (_vma->vm_end - _vma->vm_start);
         
         if(reqSize <= cl->shm.size)
         {
            sU32 reqPhysAddr;
            
            reqPhysAddr = (sU32) (_vma->vm_pgoff << PAGE_SHIFT); /* (todo) not 64bit friendly */
            
            if(reqPhysAddr == cl->shm.phys_addr)
            {
               /* (note) map CACHED memory. applications must use dsp_cache_*() functions to maintain coherency */
               
               _vma->vm_page_prot = __pgprot(pgprot_val(_vma->vm_page_prot) |
                                             (L_PTE_MT_WRITETHROUGH | L_PTE_MT_BUFFERABLE)
                                             );
               
               ///_vma->vm_page_port = pgprot_noncached(_vma->vm_page_prot);
               
               _vma->vm_flags |= (VM_RESERVED | VM_IO);
               
               /* Remap kernel memory to userspace */
               ret = remap_pfn_range(_vma,
                                     _vma->vm_start,
                                     _vma->vm_pgoff,
                                     cl->shm.size,
                                     _vma->vm_page_prot
                                     );
               if(0 == ret)
               {
                  /* Succeeded */
                  cl->shm.virt_addr = (sU32)_vma->vm_start;
                  
                  if(C64_DEBUG_LEVEL(10))
                  {
                     printk(KERN_INFO "c64::dev::map: OK, mapped CMA block to user virt_addr=0x%08x.\n",
                            cl->shm.virt_addr
                            );
                  }
                  
               }
               else
               {
                  /* Failed: remap_pfn_range() */
                  if(C64_DEBUG_LEVEL(1))
                  {
                     printk(KERN_ERR "c64::dev::map: failed to map CMA block @0x%08x to user virtual address.\n",
                            (sU32)reqPhysAddr
                            );
                  }
                  
                  ret = -EFAULT;
               }
            }
            else
            {
               /* Failed: requested physical addr does not equal CMA block start */
               if(C64_DEBUG_LEVEL(1))
               {
                  printk(KERN_ERR "c64::dev::mmap: req. physaddr (0x%08x) does not equal CMA block start (0x%08x).\n",
                         (sU32)reqPhysAddr,
                         cl->shm.phys_addr
                         );
               }
               
               ret = -EINVAL;
            }
         }
         else
         {
            /* Failed: size exceeds CMA block size */
            if(C64_DEBUG_LEVEL(1))
            {
               printk(KERN_ERR "c64::dev::mmap: req. size (%u) exceeds CMA block size (%u).\n",
                      reqSize,
                      cl->shm.size
                      );
            }
            
            ret = -EINVAL;
         }
      }
      else
      {
         /* Failed: CMA block is already mapped to user virtual address */
         if(C64_DEBUG_LEVEL(1))
         {
            printk(KERN_ERR "c64::dev::mmap: CMA block already mapped to user virt_addr=0x%08x.\n",
                   cl->shm.virt_addr
                   );
         }

         ret = -EEXIST;
      }
   }
   else
   {
      /* Failed: client not found ?!?? */
      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_ERR "c64::dev::mmap: client (filep=0x%p) not found.\n", _filep);
      }

      ret = -EBADF;
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_handle_reply() */
void dev_handle_reply(sU32 _ret1, sU32 _ret2) {

   /* (note) currently called directly from mlb IRQ handler */
   
   if(NULL != current_client)
   {
      /* Integrity check: state */
      dev_client_state_t cState = (dev_client_state_t) atomic_read(&current_client->at_state);

      if(C64_DEBUG_LEVEL(10))
      {
         printk(KERN_INFO "c64::dev_handle_reply:: recv'd msg. client=0x%08x dspCompId=0x%04x dspCmdId=0x%04x ret1=0x%08x ret2=0x%08x\n", 
                (sU32)current_client,
                current_client->c64_msg.req.fields.dsp_component_id,
                current_client->c64_msg.req.fields.dsp_cmd_id,
                _ret1,
                _ret2
                );
      }

      if(STATE_REPLY_PENDING == cState)
      {
         /* Copy return values to client message */
         current_client->c64_msg.data[0].ret = _ret1;
         current_client->c64_msg.data[1].ret = _ret2;
         
         atomic_set(&current_client->at_state, STATE_REPLY_AVAIL);
         
         /* (note) another client may already be exchanging a message when
          *         this client picks up the DSP reply.
          */
         current_client = NULL;
      }
      else
      {
         /* (note) in order to prevent kernel log spam in case sth goes terribly wrong,
          *         let's just print the error message in debug mode.
          */
         if(C64_DEBUG_LEVEL(1))
         {
            printk(KERN_ERR "c64::dev_handle_reply: unexpected client state %d.\n",
                   cState
                   );
         }
      }
   }
   else
   {
      /* received simple test message */

      if(C64_DEBUG_LEVEL(10))
      {
         printk(KERN_INFO "c64::dev_handle_reply:: recv'd test msg. ret1=0x%08x ret2=0x%08x\n", 
                _ret1,
                _ret2
                );
      }

      if(C64_DEBUG_LEVEL(1))
      {
         printk(KERN_DEBUG "c64::dev_handle_reply: no current client.\n");
      }
   }
   
}


/*--------------------------------------------------------------------------- loc_c64_device_release() */
#ifdef USE_DRIVER
static DECLARE_COMPLETION(dev_obj_is_free);

static void loc_c64_device_release(struct device *_device) {

   if(C64_DEBUG_LEVEL(10))
   {
      printk(KERN_DEBUG "c64: loc_c64_device_release() called.\n");
   }

#if 0
   complete(&dev_obj_is_free);
#endif
}
#endif /* USE_DRIVER */


/*--------------------------------------------------------------------------- dev_init() */
sBool dev_init(void) {
   sBool ret;
   int r;

   if(C64_DEBUG_LEVEL(5))
   {
      printk(KERN_INFO "c64::dev_init: ENTER\n");
   }

   current_client = NULL;

   /* Register char device 
    *
    *  (todo) alloc major number dynamically and use script to create device node
    */
#if 0
   r = register_chrdev(C64_DEV_MAJOR,
                       "c64",
                       &fops
                       );
#endif

#ifdef USE_DRIVER
   // causes crash in device_destroy() at exit
   if(0 == driver_register(&c64_driver))
#endif /* USE_DRIVER */
   {
      ////c64_driver.owner = THIS_MODULE;
#if 0
      c64_dev_nr = MKDEV(C64_DEV_MAJOR, C64_DEV_MINOR);
#endif

      r = alloc_chrdev_region(&c64_dev_nr, C64_DEV_MINOR, 1, C64_DEV_NAME);

      if(r >= 0)
      {
         c64_cdev = cdev_alloc();

         if(NULL != c64_cdev)
         {
            /* Setup device */
            c64_cdev->owner = THIS_MODULE;

            cdev_init(c64_cdev, &fops);
            
            c64_class = class_create(THIS_MODULE, "c64");

            if(! (IS_ERR(c64_class)))
            {
               c64_device = device_create(c64_class,
                                          NULL,       /* parent */
                                          c64_dev_nr,
                                          NULL,       /* drvdata */
                                          C64_DEV_NAME /*"%d",
                                                         0*/
                                          );

               if(NULL != c64_device)
               {
#ifdef USE_DRIVER
                  c64_device->release = &loc_c64_device_release;
                  c64_device->driver  = &c64_driver;
#endif /* USE_DRIVER */

#if 0
                  //mutex_lock(&c64_device->mutex); // crashes

                  r = device_bind_driver(c64_device); // crashes

                  //mutex_unlock(&c64_device->mutex);

                  if(0 == r)
#endif // 0
                  {
                     /* Initialize clients (zerofill + init atomics) */
                     loc_init_clients();
         
                     /* Add the device */
                     r = cdev_add(c64_cdev, c64_dev_nr, 1);
                     
                     if(0 == r)
                     {
                        /* Succeeded */
                        if(C64_DEBUG_LEVEL(1))
                        {
                           printk(KERN_INFO "c64: added char device dev_nr=0x%08x name=\"%s\"\n",
                                  c64_dev_nr,
                                  C64_DEV_NAME
                                  );
                        }
                        
                        ret = S_TRUE;
                     }
                     else
                     {
                        /* Failed: cdev_add() */
                        printk(KERN_ERR "c64: cdev_add() failed (r=%d)\n", r);

                        ret = S_FALSE;

                        goto failed_cdev_add;
                     }
                  }
#if 0
                  else
                  {
                     /* Failed: device_bind_driver() */
                     printk(KERN_ERR "c64: device_bind_driver() failed (r=%d).\n", r);

                     ret = S_FALSE;

                     goto failed_device_bind_driver;
                  }
#endif
               }
               else
               {
                  /* Failed: device_create() */
                  printk(KERN_ERR "c64: device_create() failed.\n");
                  
                  ret = S_FALSE;

                  goto failed_device_create;
               }
            }
            else
            {
               /* Failed: class_create() */
               printk(KERN_ERR "c64: driver_register() failed.\n");
               
               ret = S_FALSE;

               goto failed_class_create;
            }
         }
         else
         {
            /* Failed: cdev_alloc()() */
            printk(KERN_ERR "c64: cdev_alloc() failed.\n");
            
            ret = S_FALSE;

            goto failed_cdev_alloc;
         }
      }
      else
      {
         /* Failed: alloc_chrdev_region() */
         printk(KERN_ERR "c64: alloc_chrdev_region() failed (r=%d).\n", r);

         ret = S_FALSE;

         goto failed_alloc_chrdev_region;
      }
   }
#ifdef USE_DRIVER
   else
   {
      /* Failed: driver_register() */
      printk(KERN_ERR "c64: driver_register() failed.\n");

      ret = S_FALSE;
   }
#endif /* USE_DRIVER */

   goto ok;

  failed_cdev_add:
#ifdef USE_DRIVER
   device_release_driver(c64_device);

  failed_device_bind_driver:
#endif
   device_destroy(c64_class, c64_dev_nr);

  failed_device_create:
   class_destroy(c64_class);

  failed_class_create:
   cdev_del(c64_cdev);

  failed_cdev_alloc:
   unregister_chrdev_region(c64_dev_nr, 1);

  failed_alloc_chrdev_region:
#ifdef USE_DRIVER
   driver_unregister(&c64_driver);
#endif

  ok:
   if(C64_DEBUG_LEVEL(5))
   {
      printk(KERN_INFO "c64::dev_init: LEAVE ret=%d\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_exit() */
void dev_exit(void) {

   if(C64_DEBUG_LEVEL(5))
   {
      printk(KERN_INFO "c64::dev_exit: ENTER\n");
   }

   /* Teardown clients array (all file handles should have been released by now ?) */
   {
      sUI i;
      dev_client_t *cl = clients;
      
      for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
      {
         if(0/*NULL*/ != cl->shm.phys_addr)
         {
            loc_shm_free(cl);
         }
        
         /* Next client */
         cl++;
      }
   }


#if 0
   if(C64_DEBUG_LEVEL(20))
   {
      printk(KERN_DEBUG "c64::dev_exit: calling device_release_driver()\n");
   }

   device_release_driver(c64_device);
#endif

   if(C64_DEBUG_LEVEL(20))
   {
      printk(KERN_DEBUG "c64::dev_exit: calling device_destroy()\n");
   }

   device_destroy(c64_class, c64_dev_nr); // (note) crashes when driver was registered (and driver field was set)

   if(C64_DEBUG_LEVEL(20))
   {
      printk(KERN_DEBUG "c64::dev_exit: calling class_destroy()\n");
   }

   class_destroy(c64_class);

   if(C64_DEBUG_LEVEL(20))
   {
      printk(KERN_DEBUG "c64::dev_exit: calling cdev_del()\n");
   }

   cdev_del(c64_cdev);

   if(C64_DEBUG_LEVEL(20))
   {
      printk(KERN_DEBUG "c64::dev_exit: calling unregister_chrdev_region()\n");
   }

   unregister_chrdev_region(c64_dev_nr, 1);

#ifdef USE_DRIVER
   driver_unregister(&c64_driver);
#endif

#if 0
   wait_for_completion(&dev_obj_is_free);
#endif
   
   if(C64_DEBUG_LEVEL(5))
   {
      printk(KERN_INFO "c64::dev_exit: LEAVE\n");
   }
}
