/* ----
 * ---- file   : kmod.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#undef __KERNEL__
#define __KERNEL__

#undef MODULE
#define MODULE defined

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/export.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
//#include <linux/dma-contiguous.h>
#include <linux/dma-mapping.h>
#include <linux/fs.h>

#include "../include/types.h"
#include "../include/dsp_common.h"
#include "../include/msg.h"

#include "kmod.h"
#include "reg.h"
#include "mlb.h"
#include "dsp_c64.h"
#include "dev.h"


/*--------------------------------------------------------------------------- global vars */
sUI c64_debug_level = 0;


/*--------------------------------------------------------------------------- module vars */

#if 0
//----- test code -----
#include <linux/miscdevice.h>

static struct device *cma_test_device;

static ssize_t cma_test_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos) {

   ssize_t ret;
   dma_addr_t dmaAddr;
   void *virtAddr;
   struct page *pages;
   

#if 1
   printk(KERN_INFO "calling dma_alloc_coherent\n");

   virtAddr = dma_alloc_coherent(NULL,//cma_test_device,
                                 4 * 1024 * 1024,
                                 &dmaAddr,
                                 GFP_KERNEL
                                 );

   printk(KERN_INFO "dma_alloc_coherent returned virtAddr=0x%p\n", virtAddr);

   dma_free_coherent(NULL,//cma_test_device,
                     4 * 1024 * 1024,
                     virtAddr,
                     dmaAddr
                     );
#endif


#if 0
   {
      sUI numPages = ((4 * 1024 * 1024) >> 12);
      bool r;

      printk(KERN_INFO "calling dma_alloc_from_contiguous\n");
      
      pages = dma_alloc_from_contiguous(cma_test_device, 
                                        numPages,
                                        1 // alignment
                                        );

      printk(KERN_INFO "dma_alloc_from_contiguous returned pages=0x%p\n", pages);
      
      if(NULL != pages)
      {
         r = dma_release_from_contiguous(cma_test_device,
                                         pages,
                                         numPages
                                         );

         printk(KERN_INFO "dma_release_from_contiguous returned r=%d\n", r);
      }
   }
#endif




   printk(KERN_INFO "cma_test_write() LEAVE\n");

   return 0;
}

static const struct file_operations cma_test_fops = {
   .owner = THIS_MODULE,
   .write = &cma_test_write
};

static struct miscdevice cma_test_miscdevice = {
   .name = "cma_test",
   .minor = MISC_DYNAMIC_MINOR,
   .fops = &cma_test_fops
};


static int cma_test_init(void) {

   int ret = misc_register(&cma_test_miscdevice);

   if(0 == ret)
   {
      cma_test_device = cma_test_miscdevice.this_device;
      //cma_test_device->coherent_dma_mask = ~0;
   }
   else
   {
      printk(KERN_ERR "cma_test_init: failed to register device\n");
   }

   return ret;
}

static void cma_test_exit(void) {

   misc_deregister(&cma_test_miscdevice);
}
#endif


/*--------------------------------------------------------------------------- c64_init() */
static int __init c64_init(void) {
   int ret;

/*    return cma_test_init(); */


   if(C64_DEBUG_LEVEL(1))
   {
      printk(KERN_INFO "c64::c64_init: ENTER\n");
   }

   /* Put DSP into reset and disable MMU */
   dsp_init();

   /* Setup mailbox and interrupt handler */
   if(mlb_init())
   {
      /* Register "/dev/c64" char device */
      if(dev_init())
      {
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* Failed: dev_init() */
         mlb_exit();
         dsp_exit();

         ret = -1;
      }
   }
   else
   {
      /* Failed: mlb_init() */

      dsp_exit();

      ret = -1;
   }

   if(C64_DEBUG_LEVEL(1))
   {
      printk(KERN_INFO "c64::c64_init: LEAVE ret=%d\n", ret);
   }
   else
   {
      printk(KERN_INFO "c64: init ok. module build date: %s\n", __DATE__
#ifdef PANDORA_VID2_MSG_HACK
             "  (**VID2_MSG_HACK**)"
#endif /* PANDORA_VID2_MSG_HACK */
             );
   }

   return ret;
}


/*--------------------------------------------------------------------------- c64_exit() */
static void __exit c64_exit(void) {

/*    return cma_test_exit(); */

   if(C64_DEBUG_LEVEL(1))
   {
      printk(KERN_INFO "c64::c64_exit: ENTER\n");
   }

   /* Unregister char device */
   dev_exit();

   /* Disable mailbox interrupt and unregister handler */
   mlb_exit();

   /* Power off IVA2 subsystem */
   dsp_exit();

   if(C64_DEBUG_LEVEL(10))
   {
      printk(KERN_INFO "c64::c64_exit: LEAVE\n");
   }

   printk(KERN_INFO "c64: exit OK.\n");
}


/*--------------------------------------------------------------------------- c64_param_set_debug_level() */
static int c64_param_set_debug_level(const char *_val, struct kernel_param *_kp) {
   (void)_kp;

   sscanf(_val, "%u", &c64_debug_level);

   return 0;
}


/*--------------------------------------------------------------------------- module exports */
MODULE_AUTHOR("Bastian Spiegel");
MODULE_LICENSE("GPL");

MODULE_DESCRIPTION("TMS320C6000 init + ARM<>DSP messaging driver");
MODULE_SUPPORTED_DEVICE("omap3530, dm3730 (Open Pandora)");

////module_param(c64_debug_level, int, 0);
module_param_call(debug, &c64_param_set_debug_level, NULL, NULL, 0200);
MODULE_PARM_DESC(debug, "debug level (0=no debug output, 20=trace reg access)");

module_init(c64_init);
module_exit(c64_exit);
