/* ----
 * ---- file   : dsp.h
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the LESSER GNU GENERAL PUBLIC LICENSE (LGPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING_LGPL for further information.
 * ----
 * ---- info   : DSP main API. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 18Sep2013
 * ----
 * ----
 */

#ifndef __C64_TOOLS_DSP_H__
#define __C64_TOOLS_DSP_H__

#include "cplusplus_begin.h"


// If defined, actually write to the DSP registers
#define DSP_WRITE_REGS defined


/*
 * Return human-readable string for DSP error code.
 *
 */
S_EXTERN const char * S_CALL dsp_strerror(dsp_error_t _err);


/*
 * Open client connection.
 *
 *  (note) calls osal_init()
 *
 *  Returns 0 on success.
 */
S_EXTERN int dsp_open(void);


/*
 * Close client connection.
 *
 *  (note) calls osal_exit()
 *
 *  (note) calls dsp_shm_exit() in case application forgot.
 *
 */
S_EXTERN void dsp_close(void);


/*
 * Initialize GPP<>DSP shared memory.
 *
 *  (note) dsp_shm_exit() must be called before this fxn can be called again.
 *
 *  (note) numBytes will be rounded up to the next page size (4096 bytes).
 *
 *  Returns 0 on success.
 *
 */
S_EXTERN int dsp_shm_init (sU32 _numBytes);


/*
 * Query shared memory physical+virtual addresses and size.
 *
 *
 */
S_EXTERN dsp_mem_region_t dsp_shm_get (void);


/*
 * De-initialize GPP<>DSP shared memory.
 *
 *  (note) applications must ensure that this memory is not in use anymore.
 *
 *  Returns 0 on success.
 *
 */
S_EXTERN int dsp_shm_exit (void);


/*
 * Do DSP remote procedure call and wait for reply.
 *
 *  Returns 0 on success.
 *
 */
S_EXTERN int dsp_rpc(const c64_msg_t *_msgTo,
                           c64_msg_t *_msgFrom
                     );


/*
 * Do DSP remote procedure call that expects two arguments
 *  and has one return value.
 *
 *  Returns 0 on success.
 *
 */
S_EXTERN int dsp_rpc_2a_1r(sU16 _compId, 
                           sU16 _cmdId, 
                           sU32 _arg1, 
                           sU32 _arg2, 
                           sU32 *_ret
                           );


/*
 * Writeback the GPP caches for the given shared memory region.
 *
 *  (note) the given address range must be located within the GPP<>DSP shared memory area.
 *
 *  (note) dsp_shm_init() must be called before this fxn can be used.
 *
 */
S_EXTERN int dsp_cache_wb(sU32 _physAddr, sU32 _size);


/*
 * Invalidate the GPP caches for the given shared memory region.
 *
 *  (note) the given address range must be located within the GPP<>DSP shared memory area.
 *
 *  (note) dsp_shm_init() must be called before this fxn can be used.
 *
 */
S_EXTERN int dsp_cache_inv(sU32 _physAddr, sU32 _size);


/*
 * Writeback and invalidate the GPP caches for the given shared memory region.
 *
 *  (note) the given address range must be located within the GPP<>DSP shared memory area.
 *
 *  (note) dsp_shm_init() must be called before this fxn can be used.
 *
 */
S_EXTERN int dsp_cache_wbinv(sU32 _physAddr, sU32 _size);


/*
 * Query user defined DSP debug value.
 *
 *  "idx" must be in the range 0..3.
 *
 *  Returns 0 on success.
 */
S_EXTERN int dsp_debug_usr_get(sU32 _idx, sU32 *_retVal);


/*
 * Print user defined DSP debug value.
 *
 *  "idx" must be in the range 0..3.
 *
 */
S_EXTERN void dsp_debug_usr_print (sU32 _idx);


/*
 * Print c64_tools internal debug value (debugging)
 *
 *  (note) client applications should not call this fxn.
 *
 */
S_EXTERN void dsp_debug_internal_print (void);


/*
 * Call TEST ioctl (debugging).
 *
 *  (note) client applications should not call this fxn.
 *
 */
S_EXTERN int dsp_ioctl_test (void);


/*
 * Send testmessage, bypassing client handlers (debugging).
 *
 *  (note) client applications should not call this fxn.
 *
 */
S_EXTERN void dsp_send_forced_message (sU32 _val);


/*
 * To be used only by "c64_load", client applications must call dsp_close()
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_init_root (void);


/*
 * To be used only by "c64_load", client applications must call dsp_close()
 *
 */
S_EXTERN void S_CALL dsp_exit_root (void);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_mmap (sUI _physAddr, sUI _numBytes, sUI *_retAddr);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_munmap (sUI _virtAddr, sUI _numBytes);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_reg_write_32 (sUI _physAddr, sUI _value);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_reg_read_32 (sUI _physAddr, sUI *_retValue);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_reg_bit_set (sUI _physAddr, sUI _bitNr);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_reg_bit_clear (sUI _physAddr, sUI _bitNr);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_reset (void);


/*
 * To be used only by "c64_load".
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_start (sU32 _bootAddr);


/*
 * (todo) document me
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_power_down (void);


/*
 * (todo) document me
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_copy_to (sUI _physAddr, void *_srcBuf, sUI _numBytes);


/*
 * (todo) document me
 *
 */
S_EXTERN dsp_error_t S_CALL dsp_memset (sUI _physAddr, sU8 _c, sUI _numBytes);


#include "cplusplus_end.h"

#endif /* C64_TOOLS_DSP_H__ */
