/* ----
 * ---- file   : dsp.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the LESSER GNU GENERAL PUBLIC LICENSE (LGPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING_LGPL for further information.
 * ----
 * ---- info   : Arch-independent DSP utility functions. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 18Sep2013
 * ----
 * ----
 */

#define MAX_RPC_READ_RETRIES        (1000u)
#define NUM_RPC_READ_YIELD_RETRIES  (  10u)    /* yield() every "n" retries. must be >0. */

#define MAX_RPC_WRITE_RETRIES       (1000u)
#define NUM_RPC_WRITE_YIELD_RETRIES (  10u)    /* yield() every "n" retries. must be >0. */


#include <string.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* (todo) move to OSAL */
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include "../include/types.h"
#include "../include/log.h"

#include "../include/dsp_err.h"
#include "../include/dsp_common.h"
#include "../include/dsp_config.h"
#include "../include/msg.h"
#include "../include/dsp.h"
#include "../include/c64_dev.h"
#include "../include/osal_err.h"
#include "../include/osal.h"


#ifndef HAVE_OS_LINUX
#error dsp.c must be compiled for Linux (no cygwin support here)
#endif


#define USE_FAST_MMAP defined



/*--------------------------------------------------------------------------- module vars */
#ifdef USE_FAST_MMAP
static sBool b_dsp_mmap_force_real;


static dsp_mem_region_t fast_mmap[] = {

   // DSP memory (updated in dsp_init_root() with actual dsp_config)
   //   must be the first entry in this array!
   { 0,0,0 },

   // SCM / PRM DSP regs
   {
      .phys_addr = 0x48000000u,
      .size      = 0x00400000u
   }
};

#define NUM_FAST_MMAP_ENTRIES  (sizeof(fast_mmap) / sizeof(dsp_mem_region_t))

#endif // USE_FAST_MMAP


int c64_fd = -1; /* make this is a global var, just in case */

static sUI rpc_count = 0;

static c64_dev_ioctl_shm_alloc_t shm_alloc_req;


/*--------------------------------------------------------------------------- dsp_strerror() */
const char * S_CALL dsp_strerror(dsp_error_t _err) {
   switch(_err)
   {
      default:
         return NULL;

      case DSP_ERR_OK:           return "DSP_ERR_OK";

      case DSP_ERR_MEM:          return "DSP_ERR_MEM";
      case DSP_ERR_MEM_OPEN:     return "DSP_ERR_MEM_OPEN";
      case DSP_ERR_MEM_MAP:      return "DSP_ERR_MEM_MAP";
      case DSP_ERR_MEM_UNMAP:    return "DSP_ERR_MEM_UNMAP";
      case DSP_ERR_MEM_CHECK:    return "DSP_ERR_MEM_CHECK";
   }
}


/*--------------------------------------------------------------------------- dsp_open() */
int dsp_open(void) {
   int ret;
   const char *envDebug;

   /* Initialize OS abstraction layer */
   ret = osal_init();

   /* Determine log debug level */
   envDebug = getenv("C64_DEBUG");

   if(NULL != envDebug)
   {
      sscanf(envDebug, "%u", &log_debug_level);

      if(log_debug_level >= 30)
      {
         log_debug_level = 30;
      }
   }
   else
   {
      log_debug_level = 0;
   }

   /* Zerofill shared memory struct */
   memset(&shm_alloc_req, 0, sizeof(c64_dev_ioctl_shm_alloc_t));

   /* Open the c64 device */
   c64_fd = open("/dev/c64", O_RDWR | O_SYNC);

   if(-1 != c64_fd)
   {
   }
   else
   {
      log_printf(LOG_ERROR "dsp_open: failed to open \"/dev/c64\". errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
      ret = 10;
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_close() */
void dsp_close(void) {

   dsp_shm_exit();

   if(-1 != c64_fd)
   {
      /* Disconnect client */
      close(c64_fd);

      c64_fd = -1;
   }

   /* Shutdown OS abstraction layer */
   osal_exit();
}


/*--------------------------------------------------------------------------- dsp_shm_init() */
int dsp_shm_init(sU32 _numBytes) {
   int ret;
   
   if(0 == shm_alloc_req.mem.phys_addr)
   {
      shm_alloc_req.mem.size = _numBytes;

      ret = ioctl(c64_fd, C64_IOCTL_SHM_ALLOC, &shm_alloc_req);
      
      if(0 == ret)
      {
         if(LOG_DEBUG_LEVEL(30))
         {
            log_printf(LOG_INFO "dsp_shm_init: ioctl() succeeded.\n");
         }
         
         shm_alloc_req.mem.virt_addr = (sU32) mmap(0,  /* addr */
                                                   shm_alloc_req.mem.size,
                                                   (PROT_READ | PROT_WRITE),
                                                   MAP_SHARED,
                                                   c64_fd,
                                                   shm_alloc_req.mem.phys_addr
                                                   );
         
         if(shm_alloc_req.mem.virt_addr != (sU32)MAP_FAILED)
         {
            /* Succeded */
            if(LOG_DEBUG_LEVEL(1))
            {
               log_printf(LOG_INFO "dsp_shm_init: mmap() succeeded. virt_addr=0x%08x.\n",
                          shm_alloc_req.mem.virt_addr
                          );
            }
            
            ret = 0;
         }
         else
         {
            log_printf(LOG_ERROR "dsp_shm_init: mmap() failed. errno=%d (\"%s\").\n",
                       errno, strerror(errno)
                       );
            
            ret = -errno;
         }
      }
      else
      {
         /* Failed: ioctl() */
         log_printf(LOG_ERROR "dsp_shm_init: ioctl() failed. errno=%d (\"%s\").\n",
                    errno,
                    strerror(errno)
                    );
      }
   }
   else
   {
      /* Failed: shm already initialized */
      ret = 10;
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_shm_get() */
dsp_mem_region_t dsp_shm_get(void) {

   return shm_alloc_req.mem;
}


/*--------------------------------------------------------------------------- dsp_shm_exit() */
int dsp_shm_exit(void) {
   int ret;

   if(0/*NULL*/ != shm_alloc_req.mem.virt_addr)
   {
      ret = munmap((void*)shm_alloc_req.mem.virt_addr,
                   shm_alloc_req.mem.size
                   );

      if(0 != ret)
      {
         log_printf(LOG_ERROR "dsp_shm_exit: munmap() failed. virt_addr=0x%08x size=0x%08x.\n",
                    shm_alloc_req.mem.virt_addr,
                    shm_alloc_req.mem.size
                    );
      }
   }

   ret = ioctl(c64_fd, C64_IOCTL_SHM_FREE);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "dsp_shm_exit: ioctl() succeeded.\n");
      }
   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "dsp_shm_exit: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   memset(&shm_alloc_req, 0, sizeof(c64_dev_ioctl_shm_alloc_t));

   return ret;
}


/*--------------------------------------------------------------------------- dsp_ioctl_test() */
int dsp_ioctl_test(void) {
   int ret;

   ret = ioctl(c64_fd, C64_IOCTL_TEST);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "loc_ioctl_test: ioctl() succeeded.\n");
      }
   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "loc_ioctl_test: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_debug_internal_print() */
void dsp_debug_internal_print(void) {
   sU32 val;

   ssize_t numRead = read(c64_fd, &val, 4);
   
   //if(LOG_DEBUG_LEVEL(1))
   {
      log_printf(LOG_DEBUG "dsp_print_debug_internal: DSP numRead=%u, debug=0x%08x\n", (sUI)numRead, val);
   }
}


/*--------------------------------------------------------------------------- dsp_debug_usr_get() */
int dsp_debug_usr_get(sU32 _idx, sU32 *_retVal) {
   int ret;
   c64_dev_ioctl_debug_usr_t req;

   req.idx = _idx;
   
   ret = ioctl(c64_fd, C64_IOCTL_DEBUG_USR, &req);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "dsp_debug_usr_get: ioctl() succeeded val=0x%08x.\n", req.val);
      }

      if(NULL != _retVal)
      {
         *_retVal = req.val;
      }
   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "dsp_debug_usr_get: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_debug_usr_print() */
void dsp_debug_usr_print(sU32 _idx) {
   sU32 val;
   
   if(0 == dsp_debug_usr_get(_idx, &val))
   {
      log_printf(LOG_DEBUG "dsp_debug_usr_print: debug_usr[%u]=0x%08x.\n", _idx, val);
   }
}


/*--------------------------------------------------------------------------- dsp_rpc() */
int dsp_rpc(const c64_msg_t *_msgTo,
                  c64_msg_t *_msgFrom
            ) {

   /* Send RPC message to DSP */
   int ret;
   ssize_t numWritten;
   sUI retryCount = 0;

   do
   {
      numWritten = write(c64_fd, _msgTo, sizeof(c64_msg_t));

      retryCount++;

      if(0 == (retryCount % NUM_RPC_WRITE_YIELD_RETRIES))
      {
         usleep(1);
      }
   }
   while( (sizeof(c64_msg_t) != numWritten) && (EAGAIN == errno) && (retryCount < MAX_RPC_WRITE_RETRIES) );

   if(MAX_RPC_WRITE_RETRIES != retryCount)
   {
      if(sizeof(c64_msg_t) == numWritten)
      {
         /* Now read result */
         ssize_t numRead;
         retryCount = 0;
         
         do
         {
            /* (note) read() can fail with EAGAIN while DSP reply is not available, yet */
            numRead = read(c64_fd, _msgFrom, sizeof(c64_msg_t));
            
            if(LOG_DEBUG_LEVEL(30))
            {
               log_printf("loc_rpc[%08x]: [%04u] numRead=%u\n", rpc_count, retryCount, numRead);
            }
            
            retryCount++;

            if(0 == (retryCount % NUM_RPC_READ_YIELD_RETRIES))
            {
               usleep(1);
            }
         }
         while((sizeof(c64_msg_t) != numRead) && (EAGAIN == errno) && (0 != numRead) && (retryCount++ < MAX_RPC_READ_RETRIES));
         
         if(MAX_RPC_READ_RETRIES == retryCount)
         {
            log_printf(LOG_ERROR "loc_rpc[%08x]: read() timeout after %u retries.\n",
                       rpc_count,
                       MAX_RPC_READ_RETRIES
                       );
         }
         
         if(sizeof(c64_msg_t) == numRead)
         {
            if(LOG_DEBUG_LEVEL(10))
            {
               log_printf(LOG_INFO "loc_rpc[%08x]: ok, got reply from DSP. ret1=0x%08x ret2=0x%08x\n", 
                          rpc_count,
                          _msgFrom->data[0].ret,
                          _msgFrom->data[1].ret
                          );
            }
            
            /* Succeeded */
            ret = 0;
         }
         else
         {
            /* Failed: read() reply */
            log_printf(LOG_ERROR "loc_rpc[%08x]: read() failed. retryCount=%u numRead=%u errno=%d (\"%s\").\n", 
                       rpc_count,
                       retryCount,
                       numRead,
                       errno,
                       strerror(errno)
                       );
            
            dsp_debug_internal_print();
            
            ret = 10;
         }
      }
      else
      {
         /* Failed: write() */
         log_printf(LOG_ERROR "loc_rpc[%08x]: write() failed. errno=%d (\"%s\").\n", 
                    rpc_count,
                    errno,
                    strerror(errno)
                    );
         
         ret = 10;
      }
   }
   else
   {
      /* Failed: max. number of write retries exceeded */
      log_printf(LOG_ERROR "loc_rpc[%08x]: write() timeout after %u retries.\n",
                 rpc_count,
                 MAX_RPC_WRITE_RETRIES
                 );

      ret = 10;
   }

   rpc_count++;

   return ret;
}


/*--------------------------------------------------------------------------- dsp_rpc_2a_1r() */
int dsp_rpc_2a_1r(sU16 _compId, 
                  sU16 _cmdId, 
                  sU32 _arg1, 
                  sU32 _arg2, 
                  sU32 *_ret
                  ) {
   c64_msg_t msg;
   c64_msg_t reply;
   int ret;

   msg.req.fields.dsp_component_id = _compId;
   msg.req.fields.dsp_cmd_id       = _cmdId;

   msg.data[0].arg = _arg1;
   msg.data[1].arg = _arg2;

   ret = dsp_rpc(&msg, &reply);

   if(0 == ret)
   {
      if(NULL != _ret)
      {
         *_ret = reply.data[0].ret;
      }
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_cache_wb() */
int dsp_cache_wb(sU32 _physAddr, sU32 _size) {
   /* _size is #bytes */
   int ret;
   c64_dev_ioctl_cache_t req;

   req.phys_addr = _physAddr;
   req.size      = _size;
   req.action    = C64_CACHE_AC_WB;
   
   ret = ioctl(c64_fd, C64_IOCTL_CACHE, &req);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "dsp_cache_wb: ioctl() succeeded.\n");
      }
   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "dsp_cache_wb: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_cache_inv() */
int dsp_cache_inv(sU32 _physAddr, sU32 _size) {
   /* _size is #bytes */
   int ret;
   c64_dev_ioctl_cache_t req;

   req.phys_addr = _physAddr;
   req.size      = _size;
   req.action    = C64_CACHE_AC_INV;
   
   ret = ioctl(c64_fd, C64_IOCTL_CACHE, &req);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "dsp_cache_inv: ioctl() succeeded.\n");
      }

   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "dsp_cache_inv: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_cache_wbinv() */
int dsp_cache_wbinv(sU32 _physAddr, sU32 _size) {
   /* _size is #bytes */
   int ret;
   c64_dev_ioctl_cache_t req;

   req.phys_addr = _physAddr;
   req.size      = _size;
   req.action    = C64_CACHE_AC_WBINV;
   
   ret = ioctl(c64_fd, C64_IOCTL_CACHE, &req);

   if(0 == ret)
   {
      /* Succeded */
      if(LOG_DEBUG_LEVEL(30))
      {
         log_printf(LOG_INFO "dsp_cache_inv: ioctl() succeeded.\n");
      }

   }
   else
   {
      /* Failed: ioctl() */
      log_printf(LOG_ERROR "dsp_cache_inv: ioctl() failed. errno=%d (\"%s\").\n",
                 errno,
                 strerror(errno)
                 );
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_send_forced_message() */
void dsp_send_forced_message(sU32 _val) {
   
   ssize_t numWritten = write(c64_fd, &_val, 4);

   if(LOG_DEBUG_LEVEL(10))
   {
      log_printf(LOG_INFO "dsp_send_forced_message: %u bytes written.\n", numWritten);
   }
   
   if(LOG_DEBUG_LEVEL(20))
   {
      dsp_debug_internal_print();
   }
}


/*--------------------------------------------------------------------------- dsp_init_root() */
dsp_error_t dsp_init_root(void) {
   dsp_error_t ret;

   /* Initialize OS abstraction layer */
   ret = osal_init();

   if(DSP_ERR_OK == ret)
   {

#ifdef USE_FAST_MMAP
      // Init fast mmap
      {
         sUI mapIdx;

         b_dsp_mmap_force_real = S_TRUE;

         for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
         {
            fast_mmap[mapIdx].virt_addr = 0;
         }

         fast_mmap[0].phys_addr = dsp_config.reset_vector.phys_addr;
         fast_mmap[0].size      = (dsp_config.ram.phys_addr + dsp_config.ram.size) - dsp_config.reset_vector.phys_addr;

         for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
         {
            ret = dsp_mmap(fast_mmap[mapIdx].phys_addr, fast_mmap[mapIdx].size, &fast_mmap[mapIdx].virt_addr);

            if(DSP_ERR_OK == ret)
            {
               if(log_debug_level >= 10)
               {
                  log_printf(LOG_INFO "dsp_init: initialized fast_mmap[%u] physAddr=0x%08x size=0x%08x\n",
                             mapIdx,
                             fast_mmap[mapIdx].phys_addr,
                             fast_mmap[mapIdx].size
                             );
               }
            }
            else
            {
               log_printf(LOG_ERROR "dsp_init: failed to mmap() fast_mmap region phys_addr=0x%08x size=0x%08x\n",
                          fast_mmap[mapIdx].phys_addr,
                          fast_mmap[mapIdx].size
                          );
               break;
            }
            // Next entry
         }

         b_dsp_mmap_force_real = S_FALSE;
      }
#endif // USE_FAST_MMAP
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_exit_root() */
void dsp_exit_root(void) {

#ifdef USE_FAST_MMAP
   // Teardown fast mmap
   {
      sUI mapIdx;
      dsp_error_t err;
      b_dsp_mmap_force_real = S_TRUE;
      
      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if(0 != fast_mmap[mapIdx].virt_addr)
         {
            err = dsp_munmap(fast_mmap[mapIdx].virt_addr, fast_mmap[mapIdx].size);
            
            if(DSP_ERR_OK == err)
            {
               if(log_debug_level >= 10)
               {
                  log_printf(LOG_INFO "dsp_exit: de-initialized fast_mmap[%u] physAddr=0x%08x virtAddr=0x%08x size=0x%08x\n",
                             mapIdx,
                             fast_mmap[mapIdx].phys_addr,
                             fast_mmap[mapIdx].virt_addr,
                             fast_mmap[mapIdx].size
                             );
               }

            }
            else
            {
               log_printf(LOG_ERROR "dsp_exit: failed to munmap() fast_mmap region virt_addr=0x%08x size=0x%08x\n",
                          fast_mmap[mapIdx].virt_addr,
                          fast_mmap[mapIdx].size
                          );
               break;
            }
         }

         // Next entry
      }
   }
#endif // USE_FAST_MMAP

   /* Shutdown OS abstraction layer */
   osal_exit();
}


/*--------------------------------------------------------------------------- dsp_mmap() */
dsp_error_t dsp_mmap(sUI _physAddr, sUI _numBytes, sUI *_retAddr) {
   dsp_error_t ret;

   // printf("xxx dsp_mmap 1\n");

#ifdef USE_FAST_MMAP
   if(!b_dsp_mmap_force_real)
   {
      sUI mapIdx;

      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if( (_physAddr >= fast_mmap[mapIdx].phys_addr) &&
             (_physAddr <  (fast_mmap[mapIdx].phys_addr + fast_mmap[mapIdx].size))
             )
         {
            *_retAddr = (_physAddr - fast_mmap[mapIdx].phys_addr) + fast_mmap[mapIdx].virt_addr;

            if(log_debug_level >= 20)
            {
               log_printf(LOG_TRACE "**dsp_mmap(physAddr=0x%08x numBytes=0x%08x (%u))\n",
                          _physAddr,
                          _numBytes,
                          _numBytes
                          );
            }

            return DSP_ERR_OK;
         }
      }

      log_printf(LOG_WARN "dsp_mmap: fallback to mmap() physAddr=0x%08x numBytes=0x%08x\n",
                 _physAddr,
                 _numBytes
                 );
   }
#endif // USE_FAST_MMAP

   // printf("xxx dsp_mmap 2\n");

   if(log_debug_level >= 20)
   {
      // printf("xxx dsp_mmap 2a\n");
      log_printf(LOG_TRACE "dsp_mmap(physAddr=0x%08x numBytes=0x%08x (%u))\n",
                 _physAddr,
                 _numBytes,
                 _numBytes
                 );
   }

   // printf("xxx dsp_mmap 3\n");

   if(NULL != _retAddr)
   {
      
      union {
         void *ptr;
         sUI   u32;
      } virtAddr;

      sU32 physAddrAlign = DSP_PAGESIZE_ALIGN(_physAddr);
      sU32 numBytesRound = DSP_PAGESIZE_ROUND(_numBytes);

      // printf("xxx dsp_mmap 4\n");

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_mmap: physAddrAlign=0x%08x numBytesRound=0x%08x (%u)\n",
                    physAddrAlign,
                    numBytesRound,
                    numBytesRound
                    );
      }
 

      // printf("xxx dsp_mmap 5\n");

      virtAddr.ptr = osal_mmap((void*)physAddrAlign, (sSize)numBytesRound);

      // printf("xxx dsp_mmap 6 log_debug_level=%u\n", log_debug_level);

      if(NULL != virtAddr.ptr)
      {
         // Succeeded
         virtAddr.u32 = (_physAddr - physAddrAlign) + virtAddr.u32;

         ret = DSP_ERR_OK;
      }
      else
      {
         // Failed: mmap()
         log_puts(LOG_ERROR "dsp_mmap: osal_mmap() failed.\n");

         ret = DSP_ERR_MEM_MAP;
      }

      *_retAddr = virtAddr.u32;

      if(log_debug_level >= 10)
      {
         log_printf("xxx dsp_mmap() virtAddr=0x%08x\n", virtAddr.u32);
      }
   }
   else
   {
      // Failed: retAddr is NULL
      ret = DSP_ERR_MEM;
   }

   if(log_debug_level >= 20)
   {
      log_printf(LOG_TRACE "dsp_mmap: ret=%d\n", ret);
   }
   
   return ret;
}


/*--------------------------------------------------------------------------- dsp_munmap() */
dsp_error_t dsp_munmap(sUI _virtAddr, sUI _numBytes) {
   dsp_error_t ret;

#ifdef USE_FAST_MMAP
   if(!b_dsp_mmap_force_real)
   {
      sUI mapIdx;

      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if( (_virtAddr >= fast_mmap[mapIdx].virt_addr) &&
             (_virtAddr <  (fast_mmap[mapIdx].virt_addr + fast_mmap[mapIdx].size))
             )
         {
            if(log_debug_level >= 20)
            {
               log_printf(LOG_TRACE "**dsp_munmap: virtAddr=0x%08x numBytes=0x%08x (%u)\n",
                          _virtAddr,
                          _numBytes,
                          _numBytes
                          );
            }

            return DSP_ERR_OK;
         }
      }

      log_printf(LOG_WARN "dsp_munmap: fallback to munmap()\n");
   }
#endif // USE_FAST_MMAP

   if(log_debug_level >= 20)
   {
      log_printf(LOG_TRACE "dsp_munmap: virtAddr=0x%08x numBytes=0x%08x (%u)\n",
                 _virtAddr,
                 _numBytes,
                 _numBytes
                 );
   }

   {
      sUI virtAddrAlign = DSP_PAGESIZE_ALIGN(_virtAddr);
      sUI numBytesRound = DSP_PAGESIZE_ROUND(_numBytes);
      
      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_munmap: virtAddrAlign=0x%08x numBytesRound=0x%08x (%u)\n",
                    virtAddrAlign,
                    numBytesRound,
                    numBytesRound
                    );
      }

      ret = osal_munmap((void*)virtAddrAlign, numBytesRound);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_write_32() */
dsp_error_t dsp_reg_write_32(sUI _physAddr, sUI _value) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_write_32: physAddr=0x%08x oldVal=0x%08x newVal=0x%08x\n",
                    _physAddr,
                    *((sUI*)virtAddr),
                    _value
                    );
      }

      *((sUI*)virtAddr) = _value;

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_write_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_read_32() */
dsp_error_t dsp_reg_read_32(sUI _physAddr, sUI *_retValue) {
   dsp_error_t ret;
   sUI virtAddr;

   if(NULL != _retValue)
   {
      ret = dsp_mmap(_physAddr, 4, &virtAddr);
      
      if(DSP_ERR_OK == ret)
      {
         *_retValue = *((sUI*)virtAddr);

         if(log_debug_level >= 20)
         {
            log_printf(LOG_TRACE "dsp_reg_read_32: physAddr=0x%08x retVal=0x%08x\n",
                       _physAddr, *((sUI*)virtAddr)
                       );
         }
         
         // Succeeded (well, almost)
         ret = dsp_munmap(virtAddr, 4);
      }
      else
      {
         log_printf(LOG_ERROR "dsp_reg_read_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
      }
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_read_32: _retValue is NULL!!\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_bit_set() */
dsp_error_t dsp_reg_bit_set(sUI _physAddr, sUI _bitNr) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      sUI val = *((sUI*)virtAddr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_set: physAddr=0x%08x bitNr=%u oldVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      val |= (1 << _bitNr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_set: physAddr=0x%08x bitNr=%u newVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      *((sUI*)virtAddr) = val;

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_bit_set: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_bit_clear() */
dsp_error_t dsp_reg_bit_clear(sUI _physAddr, sUI _bitNr) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      sUI val = *((sUI*)virtAddr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_clear: physAddr=0x%08x bitNr=%u oldVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      val &= ~(1 << _bitNr);

      *((sUI*)virtAddr) = val;

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_clear: physAddr=0x%08x bitNr=%u newVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_bit_clear: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_copy_to() */
dsp_error_t dsp_copy_to(sUI _physAddr, void *_srcBuf, sUI _numBytes) {
   dsp_error_t ret;
   sUI virtAddr;

   if(NULL != _srcBuf)
   {
      ret = dsp_mmap(_physAddr, _numBytes, &virtAddr);
      
      if(DSP_ERR_OK == ret)
      {
         memcpy((void*)virtAddr, _srcBuf, (size_t)_numBytes);

#if 0
         sUI i;
         sU8 *d = (sU8*)virtAddr;
         sU8 *s = (sU8*)_srcBuf;

         for(i=0; i<_numBytes; i++)
         {
            sU8 sv = s[i];
            d[i] = sv;
         }
#endif
         
         // Succeeded (well, almost)
         ret = dsp_munmap(virtAddr, _numBytes);
      }
      else
      {
         log_printf(LOG_ERROR "dsp_reg_read_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
      }
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_read_32: _retValue is NULL!!\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_memset() */
dsp_error_t dsp_memset(sUI _physAddr, sU8 _c, sUI _numBytes) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, _numBytes, &virtAddr);
 
   if(DSP_ERR_OK == ret)
   {
      memset((void*)virtAddr, _c, (size_t)_numBytes);
      
      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, _numBytes);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_memset: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }
   
   return ret;
}
