/* ----
 * ---- file   : c64_tc.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : This is part of the "c64_tools" package.
 * ----
 * ---- changed: 12Sep2013, 15Sep2013, 16Sep2013, 17Sep2013, 18Sep2013
 * ----
 * ----
 */

#define NUM_BENCHMARK_ITERATIONS (100000u)
//#define NUM_BENCHMARK_ITERATIONS (10u)

//#define SHM_SIZE (4u * 1024u * 1024u)
#define SHM_SIZE (4u * 1024u * 2)
//#define SHM_SIZE (1u * 1024u * 1024u)


// Testcase enumeration:
#define TC_NONE               (0u)
#define TC_TESTMSG            (1u)  // Send 4 byte test message to DSP, bypassing client handlers
#define TC_RPC_ADD            (2u)  // Perform single remote procedure call (add two signed 32bit ints)
#define TC_RPC_ADD_BENCHMARK  (3u)  // RPC stresstest, repeatedly calls 'add' cmd of 'demo_calc' component
#define TC_IOCTL_TEST         (4u)  // Call C64_IOCTL_TEST (debugging)
#define TC_SHM_ACCESS         (5u)  // Tests shared memory access (read/write)

#define TC_NUM                (6u)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "../include/inc_libc64.h"

#include <c64p_simple_dm3730/pjt/components.h>


/*--------------------------------------------------------------------------- module vars */
#define TC_STR(a) #a
static const char *tc_names[TC_NUM] = {
   TC_STR(TC_NONE),
   TC_STR(TC_TESTMSG),
   TC_STR(TC_RPC_ADD),
   TC_STR(TC_RPC_ADD_BENCHMARK),
   TC_STR(TC_IOCTL_TEST),
   TC_STR(TC_SHM_ACCESS)
};
#undef TC_STR


static dsp_mem_region_t shm;


/*--------------------------------------------------------------------------- usage() */
static void usage(void) {
   sUI tcIdx;
   
   printf("Usage:\n");
   printf("  c64_tc [testcase_nr]\n\n");
   printf("    testcase_nr can be one of:\n");
   
   for(tcIdx=1; tcIdx < TC_NUM; tcIdx++)
   {
      printf("\t\t%02u: %s\n",
             tcIdx,
             tc_names[tcIdx]
             );
   }
   
   puts("");
}


/*--------------------------------------------------------------------------- loc_tc_testmsg() */
static void loc_tc_testmsg(void) {
   /* Send a test message bypassing the client handler in the c64 kernel module 
    *
    *  This is just to see whether the GPP<>DSP messaging in general works OK. 
    *
    *   (note) the kernel module should have been loaded with "debug=20", i.e.
    *           $ /sbin/insmod c64.ko debug=20
    */
   dsp_send_forced_message(0x37703370u);
}


/*--------------------------------------------------------------------------- loc_tc_rpc_add_test() */
static void loc_tc_rpc_add_test(void) {
   /* Test single DSP remote procedure call */
   c64_msg_t msg;
   c64_msg_t reply;
   int ret;
   
   if(LOG_DEBUG_LEVEL(1))
   {
      log_printf(LOG_DEBUG "sizeof(c64_msg_t) == %u\n", sizeof(c64_msg_t)); // must be 12 bytes
   }
   
   msg.req.fields.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
   msg.req.fields.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command
   
   msg.data[0].arg = 42 + (rand()&31);             // argument #1
   msg.data[1].arg = 22 + (rand()&31);             // argument #2
   
   ret = dsp_rpc(&msg, &reply);
   
   if(LOG_DEBUG_LEVEL(10))
   {
      dsp_debug_internal_print();
   }
   
   if(0 == ret)
   {
      log_printf(LOG_INFO "loc_rpc() OK: result=%d (expected %d)\n", 
                 *(sS32*) &reply.data[0].ret,   /* re-interpet unsigned int as signed */
                 (msg.data[0].arg + msg.data[1].arg)
                 );
      
      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }
   }
   else
   {
      log_printf(LOG_ERROR "loc_rpc() failed (ret=%d)\n", ret);
   }
}


/*--------------------------------------------------------------------------- loc_tc_rpc_add_benchmark() */
static void loc_tc_rpc_add_benchmark(void) {

   log_printf(LOG_INFO "starting benchmark\n");
   {
      c64_msg_t msg;
      c64_msg_t reply;
      sUI       tStart = osal_milliseconds_get();
      sUI       tEnd;
      sUI       tDelta;
      sUI       i;
      int       r;

      msg.req.fields.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
      msg.req.fields.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command
      
      msg.data[0].arg = 42;                           // argument #1
      msg.data[1].arg = 0;                            // argument #2

      for(i=0; i<NUM_BENCHMARK_ITERATIONS; i++)
      {
         r = dsp_rpc(&msg, &reply);

         if(0 == r)
         {
            if(reply.data[0].ret != (msg.data[0].arg + msg.data[1].arg))
            {
               log_printf(LOG_ERROR "iteration=%u: unexpected RPC result: have=0x%08x, expected=0x%08x\n",
                          i,
                          reply.data[0].ret,
                          (msg.data[0].arg + msg.data[1].arg)
                          );
               
               dsp_debug_internal_print();
               break;
            }
         }
         else
         {
            log_printf(LOG_ERROR "iteration=%u: loc_rpc() failed (r=%d).\n", r);

            dsp_debug_internal_print();
            break;
         }


         /* Next iteration */
         msg.data[0].arg++;
         msg.data[1].arg++;
      }

      if(NUM_BENCHMARK_ITERATIONS == i)
      {

         tEnd = osal_milliseconds_get();
         
         tDelta = tEnd - tStart;
         
         log_printf(LOG_INFO "benchmark finished. => %u iterations in %u millisecs.\n",
                    NUM_BENCHMARK_ITERATIONS,
                    tDelta
                    );
      }
      else
      {
         log_printf(LOG_ERROR "(benchmark aborted due to error)\n");
      }
   }
}


/*--------------------------------------------------------------------------- loc_tc_shm_access_test() */
static void loc_tc_shm_access_test(void) {
   sUI i;
   sU32 *d;
   int ret;
   sU32 chkGPP = 0;

   d = (sU32*) shm.virt_addr;

   dsp_debug_usr_print(0);

   log_printf(LOG_INFO "loc_tc_shm_access_test: virtAddr=0x%p, trying to write..\n", d);

   for(i=0; i<(shm.size >> 2); i++)
   {
      d[i] = i;
      chkGPP += i;
   }

   log_printf(LOG_INFO "loc_tc_shm_access_test: OK, wrote %d dwords (%u kBytes).\n", i, (i >> 8));

   /* Writeback GPP caches so DSP can see the updated data */
   ret = dsp_cache_wb(shm.phys_addr,
                      shm.size
                      );

   if(0 == ret)
   {
      /* Now verify that DSP sees the same data by calling the DSP-side demo_checksum::add() method */
      sU32 chkDSP = 0xCCccCCccu;

      ret = dsp_rpc_2a_1r(COMPONENT_DEMO_CHECKSUM,
                          DEMO_CHECKSUM_ADD,
                          shm.phys_addr,
                          shm.size,
                          &chkDSP
                          );

      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }

      if(chkDSP == chkGPP)
      {
         log_printf(LOG_INFO "loc_tc_shm_access_test: OK, DSP+GPP checksums match (chk=0x%08x)\n",
                    chkDSP
                    );
      }
      else
      {
         log_printf(LOG_ERROR "loc_tc_shm_access_test: failed, checksum mismatch (gpp=0x%08x dsp=0x%08x)\n",
                    chkGPP,
                    chkDSP
                    );

         dsp_debug_usr_print(0);
      }
   }
   else
   {
      log_printf(LOG_ERROR "loc_tc_shm_access_test: loc_cache_wb() failed (ret=%d)\n", ret);
   }
}


/*--------------------------------------------------------------------------- main() */
int main(int argc, char**argv) {
   int ret;

   if(1 == argc)
   {
      usage();

      return 0;
   }

   /* Open client connection */
   ret = dsp_open();

   if(0 == ret)
   {
      sUI tcIdx = TC_NONE;

      /* Determine selected testcase */
      if(2 == argc)
      {
         sscanf(argv[1], "%u", &tcIdx);

         if(tcIdx >= TC_NUM)
         {
            log_printf(LOG_ERROR "invalid testcase idx %u\n", tcIdx);

            tcIdx = TC_NONE;
         }
      }

      log_printf(LOG_INFO "selected testcase %u (\"%s\")\n", tcIdx, tc_names[tcIdx]);

      if(LOG_DEBUG_LEVEL(10))
      {
         dsp_debug_internal_print();
      }

      ret = 0;

      /* Allocate contiguous memory block */
      if(0 == dsp_shm_init(SHM_SIZE))
      {
         /* Query shared memory attributes */
         shm = dsp_shm_get();


         /* Run testcase */
         switch(tcIdx)
         {
            default:
               break;
               
            case TC_TESTMSG:
            {
               loc_tc_testmsg();
            }
            break;
            
            case TC_RPC_ADD:
            {
               loc_tc_rpc_add_test();
            }
            break;
            
            
            case TC_RPC_ADD_BENCHMARK:
            {
               /* Benchmark DSP remote procedure call speed */
               
               loc_tc_rpc_add_benchmark();
               
            }
            break;
            
            case TC_IOCTL_TEST:
            {
               dsp_ioctl_test();
            }
            break;
            
            case TC_SHM_ACCESS:
            {
               loc_tc_shm_access_test();
            }
            break;
            
         } // switch

         /* Free contiguous memory block */
         dsp_shm_exit();
         
      } /* if dsp_shm_init() */
         
      /* Disconnect client */
      dsp_close();

   } /* if dsp_open() */

   return ret;
}
